#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# ====#====#====#====
# Author: jys
# File: httpClient
# Time: 2019/3/12
# ====#====#====#====


from requests.exceptions import ReadTimeout, ConnectTimeout, HTTPError
import requests
import logging

_logger = logging.getLogger(__name__)
HTTP_WAIT_TIME = 2


class httpClientException(Exception):
	'''
		base httpClient exception
	'''

	def __init__(self, msg="", value=1):
		'''
		Initialize the exception
		:param value: the err code
		:param msg: the err msg
		'''

		self.value = value
		self.msg = msg

	def __str__(self):
		'''
		return the exception message
		:return str:
		'''
		return "Err: %s ErrCode: %s" % (str(self.msg), self.value)

	__repr__ = __str__


class HttpCommException(httpClientException):
	''' http comm error '''

	def __init__(self, msg="", value=1):
		'''Initialize the exception
		:param value: the err code
		:param msg: the err msg
		'''

		message = "[HttpComm] %s" % (str(msg))
		httpClientException.__init__(self, message, value)


class CertManger(object):
	def __init__(self, ca=None, cert=None, key=None):
		self.ca = ca
		self.cert = cert
		self.key = key


class HttpClient(object):

	def __init__(self, hostUrl, clientCerts=None):
		'''
			httpCLient 客户端类
		:param hostUrl: 连接地址
		:param clientCerts: 证书
		'''
		sslFlag = False
		self.session = requests.Session()

		if (None != clientCerts):
			if (isinstance(clientCerts, CertManger)):
				raise TypeError

			self.session.cert = (clientCerts.cert, clientCerts.key)
			self.session.verify = clientCerts.ca
			sslFlag = True
			
			
		if(hostUrl.startswith("http")):
			self._hostUrl = hostUrl

		else:
			if (sslFlag):
				self._hostUrl = "https://" + hostUrl
			else:
				self._hostUrl = "http://" + hostUrl

	def httpGetJson(self, url, jsonData):
		'''
			http get 方式, 操作json 数据
		:param url: 相对路径
		:param jsonData:  dict 类型的数据
		:return: dict 类型数据
		'''

		headers = {"Content-Type": "application/json"}
		try:
			_logger.debug("-----> url: %s, httpGetJson In args: %s", url, jsonData)
			response = self.session.get(url= self._hostUrl + url, headers=headers, params=jsonData,
										timeout=HTTP_WAIT_TIME)
			responseData = response.json()

			_logger.debug("<----- httpGetJson output params: %s", responseData)
			return responseData

		except (ReadTimeout, ConnectTimeout, HTTPError) as err:
			_logger.error('%s', err)
			_logger.exception("Err")

		except:
			_logger.exception("Err")

		raise HttpCommException('client http trans err')

	def httpPostJson(self, url, jsonData):
		'''
			http post, 操作json数据
		:param url: 相对路劲
		:param jsonData: dict 类型的数据
		:return: dict 数据类型
		'''

		headers = {"Content-Type": "application/json"}

		try:
			_logger.debug("-----> url: %s, httpPostJson In args: %s", url, jsonData)
			response = self.session.post(url= self._hostUrl + url, headers=headers, json=jsonData,
										 timeout=HTTP_WAIT_TIME)
			responseData = response.json()

			_logger.debug("<----- httpPostJson output params: %s", responseData)
			return responseData

		except (ReadTimeout, ConnectTimeout, HTTPError) as err:
			_logger.exception('Err: %s', err)

		except:
			_logger.exception('Err: Post data to %s error.' % url)

		raise HttpCommException('client http trans err')

	def httpPostJsonWithFile(self, url, jsonData, fileInfo):
		'''
			http post,发送json 报文,并上传文件
		:param url: 相对路劲
		:param jsonData:
		:param fileInfo[dict]:   eg: {fileKey, fileName}
		:return:
		'''

		_logger.debug("-----> url: %s, httpPostJsonWithFile In args: %s", url, jsonData)

		files = {}
		for fileKey, fileName in fileInfo.items():
			files[fileKey] = open(fileName, "rb")

		try:
			response = self.session.post(url= self._hostUrl + url, data=jsonData, files=files, timeout=HTTP_WAIT_TIME)
			responseData = response.json()

			_logger.debug("<----- httpPostJsonWithFile output params: %s", responseData)
			return responseData

		except (ReadTimeout, ConnectTimeout, HTTPError) as err:
			_logger.error('%s', err)
			_logger.exception("Err")

		except:
			_logger.exception("Err")

		raise HttpCommException('client http trans err')


if (__name__ == "__main__"):
	format = '%(asctime)s-%(threadName)-9s-%(filename)s[line:%(lineno)d]-%(levelname)s: %(message)s'
	logging.basicConfig(level=logging.INFO, format=format)

	obj = ParkingHttpClient("http://localhost:1234")
	obj.debug()
