local const = require("const")
local json = require("json")
local utils = require("utils2")
local ngx_re_match = ngx.re.match

local function_pattern = [[^(acs:fc:([a-z]{2}-[a-z]+(-\d)?)?:(\d{16})?:)?service:([a-zA-Z_][\w-]{0,127}):function:([a-zA-Z_][\w-]{0,127})$]]
local error_type_internal = 0
local error_type_unhandled = 1
local error_type_handled = 2

utils.check_auth()
utils.check_permission("FunctionComputeManage")

local request_method = ngx.var.request_method
if request_method ~= "POST" then
    return utils.error_exit(ngx.HTTP_NOT_ALLOWED, "request method not allowed: " .. request_method)
end

local function invoke_function(func)
    local args = {}
    args["serviceName"] = func.service or ""
    args["functionName"] = func.func_name or ""
    args["regionId"] = func.region or ""
    args["accountId"] = func.account_id or ""
    local payload = func.payload or ""
    args["payload"] = ngx.encode_base64(payload)

    local ok, ret, err = utils.dbus_call_edge(
            const.function_compute.wkn,
            const.function_compute.obj,
            const.function_compute.inf,
            "invokeFunction", json.encode(args))
    if not ok then
        return false, error_type_internal, err
    end

    local res
    ok, res = pcall(json.decode, ret[1])
    if not ok or type(res) ~= "table" then
        return false, error_type_internal, "response from upstream service not a JSON object"
    end

    local code = res["code"]
    local data = res["data"]
    if code == 101505 then
        data = ngx.decode_base64(data)
        if not data then
            return false, error_type_internal, "response data from upstream service not base64 encoded"
        end
        return false, error_type_handled, data
    end
    if code ~= 0 then
        return false, error_type_unhandled, res["message"]
    end
    if type(data) == "string" then
        data = ngx.decode_base64(data)
        if not data then
            return false, error_type_internal, "response data from upstream service not base64 encoded"
        end
    end
    return true, data
end


local function_arn = ngx.var[1]
local items = ngx_re_match(function_arn, function_pattern)
if not items then
    return utils.error_exit(ngx.HTTP_BAD_REQUEST, "invalid function ARN")
end

ngx.req.read_body()
local body = ngx.req.get_body_data()
if body ~= nil and type(body) ~= "string" then
    return utils.error_exit(ngx.HTTP_BAD_REQUEST, "Payload not a JSON")
end

local ok, res, err = invoke_function({
    service = items[5] or "",
    func_name = items[6] or "",
    region = items[2] or "",
    account_id = items[4] or "",
    payload = body,
})

local resp = {}

-- Handle errors.
if not ok then
    if res == error_type_internal then
        return utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, err)
    elseif res == error_type_unhandled then
        ngx.header["X-Fc-Error-Type"] = "Unhandled"
        resp["Message"] = err
        ngx.say(json.encode(resp))
        return
    else
        ngx.header["X-Fc-Error-Type"] = "Handled"
        resp["Message"] = ""
        resp["StackTrace"] = ""
        local ok, data = pcall(json.decode, err)
        if ok then
            resp["Message"] = data["errorMessage"]
            resp["StackTrace"] = data["stackTrace"]
        end
        ngx.say(json.encode(resp))
        return
    end
end

-- Handle data from the function.
ngx.say(res)
